#ifndef EDIT_DISTANCE_INC
#define EDIT_DISTANCE_INC

#include "general.h"
#include "Object.h"
#include "Dictionary.h"

/** This class can be used to compute the "edit distance" between
    two sequences. It computes the number of insertions, deletions and
    substitutions. The overall distance is the sum of these numbers
    weighted by their cost (which are intergers equal to 1 by default).

    @author Samy Bengio (bengio@idiap.ch)
*/
class EditDistance : public Object
{
  public:
    /// the total edit distance between two sequences
    real accuracy;
    /// the number of insertions (weighted by their cost)
    int n_insert;
    /// the number of deletions (weighted by their cost)
    int n_delete;
    /// the number of substitutions (weighted by their cost)
    int n_subst;
    /// the number of sequences measured (used to normalize #dist# by #n_seq#)
    int n_seq;
    /// the cost of one insertion
    int insert_cost;
    /// the cost of one deletion
    int delete_cost;
    /// the cost of one substitution
    int subst_cost;

    /** optionally, if a dictionary is given, it is used to print the
        obtained and desired sequence using the dictionary translation
    */
    Dictionary* dict;

    /// the obtained sequence
    int* obtained;
    /// the size of the obtained sequence
    int obt_size;

    /// the desired sequence
    int* desired;
    /// the size of the desired sequence
    int des_size;

    ///
    EditDistance();

    /// sets the different costs
    virtual void setCosts(int i_cost, int d_cost, int s_cost);

    /** computes the edit distance between #obtained# and #desired#.
        if #dict_# is given, it is used to print the sequences.
    */
    virtual void distance(int* obtained, int obt_size, int* desired, int des_size, Dictionary* dict_ = NULL);

    /// accumulates the distances of the current object and the given object
    virtual void add(EditDistance* d);

    /// prints the edit distance and optionally the sequences
    virtual void print(FILE *f,bool print_sequences=false);

    /** prints the edit distance ratio (divided by the number of sequnces) 
        and optionally the obtained and desired sequences
    */
    virtual void printRatio(FILE *f);

    virtual void reset();
    ~EditDistance();
};

#endif
