
#include "EigenDiagonalGMM.h"
#include "log_add.h"
#include "random.h"
#include "matrix.h"

EigenDiagonalGMM::EigenDiagonalGMM(int n_observations_, int n_gaussians_, real* var_threshold_, real prior_weights_,real** eigen_vect_,int eigen_dimension_,real weight_on_prior_,DiagonalGMM* prior_distribution_) : DiagonalGMM(n_observations_,n_gaussians_,var_threshold_,prior_weights_)
{
	w = NULL;
	eigen_vect = eigen_vect_;
	eigen_dimension = eigen_dimension_;
	Q = NULL;
	pre_Q = NULL;
	pre_v = NULL;
	v = NULL;
	prior_distribution = prior_distribution_;	
	weight_on_prior = weight_on_prior_;
}

void EigenDiagonalGMM::allocateMemory()
{
	DiagonalGMM::allocateMemory();
	w = (real*)xalloc(sizeof(real)*eigen_dimension);
	v = (real*)xalloc(sizeof(real)*eigen_dimension);
  Q = (real**)xalloc(sizeof(real*)*eigen_dimension);
	pre_Q = (real***)xalloc(sizeof(real**)*eigen_dimension);
	pre_v = (real***)xalloc(sizeof(real**)*eigen_dimension);
	real** q_i = Q;
	real*** pre_Q_i = pre_Q;
	real*** pre_v_i = pre_v;
	for(int i=0;i<eigen_dimension;i++){
		*q_i++ = (real*)xalloc(sizeof(real)*eigen_dimension);
		*pre_Q_i = (real**)xalloc(sizeof(real*)*eigen_dimension);
		real** pre_Q_j = *pre_Q_i++;
		for(int j=0;j<eigen_dimension;j++)
			*pre_Q_j++ = (real*)xalloc(sizeof(real)*n_gaussians);
		*pre_v_i = (real**)xalloc(sizeof(real*)*n_gaussians);
		real** pre_v_j = *pre_v_i++;
		for(int j=0;j<n_gaussians;j++)
			*pre_v_j++ = (real*)xalloc(sizeof(real)*n_observations);
	}
}

void EigenDiagonalGMM::freeMemory()
{
	if(w){
		free(w);
		w = NULL;
	}
	if(v){
		free(v);
		v = NULL;
	}
	if(Q){
		real** tmp = Q;
		for(int i=0;i<eigen_dimension;i++)
			free(*tmp++);
		free(Q);
		Q = NULL;
	}
	if(pre_Q){
		for(int i=0;i<eigen_dimension;i++){
			for(int j=0;j<eigen_dimension;j++)
				free(pre_Q[i][j]);
			free(pre_Q[i]);
		}
		free(pre_Q);
		pre_Q = NULL;
	}


	if(pre_v){
		for(int i=0;i<eigen_dimension;i++){
			for(int j=0;j<n_gaussians;j++)
				free(pre_v[i][j]);
			free(pre_v[i]);
		}
		free(pre_v);
		pre_Q = NULL;
	}
}

int EigenDiagonalGMM::numberOfParams()
{
  return (n_observations * n_gaussians * 2) // number of means and var
        + n_gaussians; // number of weights;
}

void EigenDiagonalGMM::reset()
{
	 copyList(params,prior_distribution->params);

	 //pre compute some value for Q: pre_Q
	 real*** pre_q_i = pre_Q;
	 real** eigen_vect_i = eigen_vect;
	 for(int i=0;i<eigen_dimension;i++,eigen_vect_i++){
		 real** pre_q_j = *pre_q_i++;
		 real** eigen_vect_j = eigen_vect;
		 for(int j=0;j<eigen_dimension;j++){
			 real* eigen_vect_i_kl = *eigen_vect_i;
			 real* eigen_vect_j_kl = *eigen_vect_j++;
			 real** var_k = var;
			 real* pre_q_k = *pre_q_j++;
			 for(int k=0;k<n_gaussians;k++){
				 real *var_l = *var_k++;
				 real res1 = 0.0;
				 for(int l=0;l<n_observations;l++){
					 res1 += *eigen_vect_j_kl++ / *var_l++ * *eigen_vect_i_kl++;	
				 }
				 *pre_q_k++ = res1;
			 }
		 }
	 }
	 real*** pre_v_i = pre_v;
	eigen_vect_i = eigen_vect;
	for(int i=0;i<eigen_dimension;i++){
		real** var_j = var;
		real* eigen_vect_jk = *eigen_vect_i++;
		real** pre_v_j = *pre_v_i++;
		for(int j=0;j<n_gaussians;j++){
			real* var_k = *var_j++;
			real* pre_v_k = *pre_v_j++;
			for(int k=0;k<n_observations;k++)
			 	*pre_v_k++ = *eigen_vect_jk++ / *var_k++;
		}
	}


}

void EigenDiagonalGMM::frameEMAccPosteriors(real *observations, real log_posterior, real *inputs, int t)
{
  real log_prob = log_probabilities[t];
  real *p_weights_acc = weights_acc;
  real *lp_i = log_probabilities_g[t];
  real *log_w_i = log_weights;
  for (int i=0;i<n_gaussians;i++) {
    real post_i = exp(log_posterior + *log_w_i++ + *lp_i++ - log_prob);
    *p_weights_acc++ += post_i;
    real* means_acc_i = means_acc[i];
    real *x = observations;
    for(int j = 0; j < n_observations; j++) {
      *means_acc_i++ += post_i * *x++;
    }
  }
}

void EigenDiagonalGMM::eMUpdate()
{
	//compute matrix v
	/*real* v_i = v;
	real** eigen_vect_i = eigen_vect;
	for(int i=0;i<eigen_dimension;i++,v_i++){
		real** means_acc_j = means_acc;
		real** var_j = var;
		real* eigen_vect_jk = *eigen_vect_i++;
		*v_i = 0.0;
		for(int j=0;j<n_gaussians;j++){
			real* means_acc_k = *means_acc_j++;
			real* var_k = *var_j++;
			for(int k=0;k<n_observations;k++)
				*v_i += *means_acc_k++ * *eigen_vect_jk++ / *var_k++;
		}
	}*/
	real* v_i = v;
	real*** pre_v_i = pre_v;
	for(int i=0;i<eigen_dimension;i++,v_i++){
		real** means_acc_j = means_acc;
		*v_i = 0.0;
		real** pre_v_j = *pre_v_i++;
		for(int j=0;j<n_gaussians;j++){
			real* means_acc_k = *means_acc_j++;
			real* pre_v_k = *pre_v_j++;
			for(int k=0;k<n_observations;k++)
				*v_i += *means_acc_k++ * *pre_v_k++;
		}
	}

	
/*
	for (int i=0;i<eigen_dimension;i++) {
		v[i] = 0;
		for (int j=0;j<n_gaussians;j++) {
			for (int k=0;k<n_observations;k++) {
				v[i] += means_acc[j][k] * eigen_vect[i][j*n_observations+k] / var[j][k];
			}
	  }
  }
*/
/*	//compute matrix Q
	real** q_i = Q;
	eigen_vect_i = eigen_vect;
	for(int i=0;i<eigen_dimension;i++,eigen_vect_i++){
		real* w_j = w;
		real* q_j = *q_i++;
		real** eigen_vect_j = eigen_vect;
		for(int j=0;j<eigen_dimension;j++,q_j++,w_j++){
			real* eigen_vect_i_kl = *eigen_vect_i;
			real* weights_acc_k = weights_acc;
			real* eigen_vect_j_kl = *eigen_vect_j++;
			real** var_k = var;
			*q_j = 0.0;
			for(int k=0;k<n_gaussians;k++){
				real res =  *weights_acc_k++;
				real *var_l = *var_k++;
				real res1 = 0.0;
				for(int l=0;l<n_observations;l++){
					res1 += *eigen_vect_j_kl++ / *var_l++ * *eigen_vect_i_kl++;	
				}
				*q_j += res*res1;
			}
		}
	}*/


	//compute matrix Q
	real** q_i = Q;
	real*** pre_q_i = pre_Q;
	for(int i=0;i<eigen_dimension;i++){
		real* q_j = *q_i++;
		real** pre_q_j = *pre_q_i++;
		for(int j=0;j<eigen_dimension;j++,q_j++){
			real* weights_acc_k = weights_acc;
			*q_j = 0.0;
			real* pre_q_k = *pre_q_j++;
			for(int k=0;k<n_gaussians;k++){
				real res =  *weights_acc_k++;
				*q_j += res* *pre_q_k++;
			}
		}
	}
	
	/*
	for (int i=0;i<eigen_dimension;i++) {
	  for (int j=0;j<eigen_dimension;j++) {
			Q[i][j] = 0;
		  for (int k=0;k<n_gaussians;k++) {
			  for (int l=0;l<n_observations;l++) {
				  Q[i][j] += weights_acc[k] * eigen_vect[j][k*n_observations+l] * eigen_vect[i][k*n_observations+l] / var[k][l];
			  }
			}
	  }
	}
*/
	//solve equation v = Q * w => w = Q^-1 * v
	Perm* pivot = new Perm(eigen_dimension);	
	Mat* q_mat = new Mat(Q,eigen_dimension,eigen_dimension);
	Vec* v_vec = new Vec(v,eigen_dimension);
	Vec* w_vec = new Vec(w,eigen_dimension);
	mxLUFactor(q_mat, pivot);
	mxLUSolve (q_mat, pivot, v_vec,w_vec);
	delete(pivot);
	delete(q_mat);
	delete(v_vec);
	delete(w_vec);
	//reset means
	real** means_i = means;
	for(int i=0;i<n_gaussians;i++){
		real* means_j = *means_i++;
		for(int j=0;j<n_observations;j++){
			*means_j++ = 0.0;
		}
	}

	//update with means = P * w
	real** eigen_vec_h = eigen_vect;
	real* w_h = w;
	for(int h=0;h<eigen_dimension;h++,w_h++){
		real* eigen_vec_ij = *eigen_vec_h++;
		real** means_i = means;
		for(int i=0;i<n_gaussians;i++){
			real* means_j = *means_i++;
			for(int j=0;j<n_observations;j++){
				*means_j++ += *eigen_vec_ij++ * *w_h;
			}
		}
	}
/*
	for (int i=0;i<n_gaussians;i++) {
		for (int j=0;j<n_observations;j++) {
			means[i][j] = 0;
			for (int k=0;k<eigen_dimension;k++) {
				means[i][j] += eigen_vect[k][i*n_observations+j] * w[k];
			}
		}
	}
	*/
		for (int i=0;i<n_gaussians;i++) 
			for (int j=0;j<n_observations;j++) {
				means[i][j] = weight_on_prior * prior_distribution->means[i][j] + (1-weight_on_prior) * means[i][j]; 
			}
}


EigenDiagonalGMM::~EigenDiagonalGMM()
{
  freeMemory();
}
