#include "MAPDiagonalGMM.h"
#include "log_add.h"
#include "random.h"

MAPDiagonalGMM::MAPDiagonalGMM(int n_observations_, int n_gaussians_, real* var_threshold_, real prior_weights_, DiagonalGMM* prior_distribution_,real weight_on_prior_,real thrd_) : DiagonalGMM(n_observations_,n_gaussians_,var_threshold_,prior_weights_)
{
  prior_distribution = prior_distribution_;
	thrd = thrd_;
  weight_on_prior = weight_on_prior_;
}

void MAPDiagonalGMM::reset()
{
  // here, initialize the parameters to the parameters of the prior
  // distribution
  if (prior_distribution)
    copyList(params,prior_distribution->params);
  else
    DiagonalGMM::reset();

}

void MAPDiagonalGMM::frameViterbiAccPosteriors(real *observations, real log_posterior, real *inputs, int t)
{
  real *lp_i = log_probabilities_g[t];
  real *log_w_i = log_weights;
  int best = -1;
  real best_sc = -FLT_MAX;
  for (int i=0;i<n_gaussians;i++,lp_i++,log_w_i++) {
		real score = *lp_i + *log_w_i;
		if(score > best_sc){
		   best = i;
			 best_sc = score;
		}
	}
	real post_i = exp(log_posterior);
	weights_acc[best] += post_i; 
	real* means_acc_i = means_acc[best];
	real *x = observations;
	for(int j = 0; j < n_observations; j++) {
		*means_acc_i++ += post_i * *x++;
	}
}


void MAPDiagonalGMM::frameEMAccPosteriors(real *observations, real log_posterior, real *inputs, int t)
{
  real log_prob = log_probabilities[t];
  real *p_weights_acc = weights_acc;
  real *lp_i = log_probabilities_g[t];
  real *log_w_i = log_weights;
  for (int i=0;i<n_gaussians;i++) {
    real post_i = exp(log_posterior + *log_w_i++ + *lp_i++ - log_prob);
    *p_weights_acc++ += post_i;
    real* means_acc_i = means_acc[i];
    real *x = observations;
    for(int j = 0; j < n_observations; j++) {
      *means_acc_i++ += post_i * *x++;
    }
  }
}

void MAPDiagonalGMM::eMUpdate()
{
  // just the means
  real* p_weights_acc = weights_acc;
  for (int i=0;i<n_gaussians;i++,p_weights_acc++) {
    if (*p_weights_acc < thrd) {
      warning("Gaussian %d of GMM is not used in EM: nobs: %.3f",i,*p_weights_acc);
    } else {
      real* p_means_prior_i = prior_distribution->means[i];
      real* p_means_i = means[i];
      real* p_means_acc_i = means_acc[i];
      for (int j=0;j<n_observations;j++) {
        *p_means_i++ = weight_on_prior * *p_means_prior_i++ +
          (1 - weight_on_prior) * *p_means_acc_i++ / *p_weights_acc;
      }
    }
  }
}

void MAPDiagonalGMM::eMIterInitialize()
{
  // initialize the accumulators to 0 and compute pre-computed value
  for (int i=0;i<n_gaussians;i++) {
    real *pm = means_acc[i];
    real *ps = var_acc[i];
    real *v = var[i];
    real *sum = &sum_log_var_plus_n_obs_log_2_pi[i];
    *sum = n_observations * LOG_2_PI;
    real *mh_i = minus_half_over_var[i];
    for (int j=0;j<n_observations;j++) {
      *pm++ = 0.;
      *ps++ = 0.;
      *mh_i++ = -0.5 / *v;
      *sum += log(*v++);
    }
    *sum *= -0.5;
    weights_acc[i] = prior_weights;
  }
}

void MAPDiagonalGMM::iterInitialize()
{
  for (int i=0;i<n_gaussians;i++) {
    real* means_prior_i = prior_distribution->means[i];
    real* means_i = means[i];
    for (int j=0;j<n_observations;j++,means_i++,means_prior_i++) {
      *means_i = weight_on_prior * *means_prior_i +
          (1 - weight_on_prior) * *means_i;
    }
  }
}

void MAPDiagonalGMM::frameBackward(real *observations, real *alpha, real *inputs, int t)
{
  real log_prob = log_probabilities[t];
  real *lp_i = log_probabilities_g[t];
  real *lw = log_weights;
  for (int i=0;i<n_gaussians;i++,lw++,lp_i++) {
    real post_i =  - *alpha * exp(*lw + *lp_i - log_prob);
    real* obs = observations;
    real* means_i = means[i];
    real* dmeans_i = dmeans[i];
    real* var_i = var[i];
    for (int j=0;j<n_observations;j++,var_i++,obs++,means_i++,dmeans_i++) {
      real xmuvar = (*obs - *means_i) / *var_i;
      real dm = post_i * 2. * xmuvar;
      *dmeans_i += dm;
    }
  }
}

MAPDiagonalGMM::~MAPDiagonalGMM()
{
}
